import argparse
import json
import sys


def gen_code_cov_badge_cmd(args):
    input_filepath = args.input_file
    output_filepath = args.output_file

    def get_coverage_color(percentage) -> str:
        color_map = {0: "red", 50: "orange", 80: "yellow", 90: "green"}
        coverage_level = sum(level <= percentage for level in color_map) - 1
        color_map_key = list(color_map.keys())[coverage_level]
        return color_map[color_map_key]

    with open(input_filepath) as in_f:
        data = json.load(in_f)
        percentage = data["totals"]["percent_covered"]
        percent_display = data["totals"]["percent_covered_display"]
        color = get_coverage_color(percentage)
        coverage_badge_config = f"""{{
    "schemaVersion": 1,
    "label": "code coverage",
    "message": "{percent_display}%",
    "color": "{color}"
}}"""
        with open(output_filepath, "w") as out_f:
            out_f.write(coverage_badge_config)


def main():
    parser = argparse.ArgumentParser(
        description="CLI tool to perform utility tasks in repository",
    )

    subparsers = parser.add_subparsers()
    cov_badge_parser = subparsers.add_parser(
        "gen-cov-badge",
        description="Generate code coverage badge config",
    )
    cov_badge_parser.add_argument("input_file", type=str)
    cov_badge_parser.add_argument("output_file", type=str)
    cov_badge_parser.set_defaults(func=gen_code_cov_badge_cmd)
    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError:
        parser.print_help()
        sys.exit(0)


if __name__ == "__main__":
    main()



import json
import logging
import math
import os
import sys
import tempfile
from collections.abc import Callable, Iterable, Sequence
from enum import Enum
from typing import Any, Literal, TextIO, cast

from geojson_pydantic import (
    Feature,
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from geojson_pydantic.geometries import Geometry
from geojson_pydantic.types import LineStringCoords, Position
from pydantic import BaseModel
from pyproj import CRS
from shapely import LineString as ShpLineString
from shapely import Point as ShpPoint

from geodense.geojson import CrsFeatureCollection
from geodense.models import DEFAULT_PRECISION_METERS, DenseConfig, GeodenseError
from geodense.types import (
    GeojsonCoordinates,
    GeojsonGeomNoGeomCollection,
    GeojsonObject,
    Nested,
    ReportLineString,
    T,
)

TWO_DIMENSIONAL = 2
THREE_DIMENSIONAL = 3
DEFAULT_CRS_2D = "OGC:CRS84"
DEFAULT_CRS_3D = "OGC:CRS84h"
SUPPORTED_FILE_FORMATS = {
    "GeoJSON": [".geojson", ".json"],
}

logger = logging.getLogger("geodense")


def densify_geojson_object(geojson_obj: GeojsonObject, dc: DenseConfig) -> None:
    _geom_type_check(geojson_obj, "densify")
    geom_densify_fun = _get_geom_densify_fun(dc)
    _ = apply_function_on_geojson_geometries(geojson_obj, geom_densify_fun)


def density_check_geojson_object(
    geojson_obj: GeojsonObject, dc: DenseConfig
) -> CrsFeatureCollection:
    _geom_type_check(geojson_obj, "density-check")
    density_check_fun = get_density_check_fun(dc)
    result: Nested[ReportLineString] = apply_function_on_geojson_geometries(
        geojson_obj, density_check_fun
    )
    flat_result: list[
        ReportLineString
    ] = list(  # filter out None values, these occur when point geometries are part of input
        filter(lambda x: x is not None, flatten(result))
    )
    report_fc = report_line_string_to_geojson(
        flat_result, ":".join(dc.src_crs.to_authority())
    )
    return report_fc


def get_density_check_fun(
    densify_config: DenseConfig,
) -> Callable:
    def density_check(
        geometry: GeojsonGeomNoGeomCollection,
    ) -> Nested[ReportLineString]:
        return check_density_geometry_coordinates(geometry.coordinates, densify_config)

    return density_check


def _validate_dependent_file_args(
    input_file_path: str,
    output_file_path: str | None = None,
    overwrite: bool = False,
) -> None:
    if output_file_path is not None and (
        input_file_path == output_file_path and input_file_path != "-"
    ):
        raise GeodenseError(
            f"input_file and output_file arguments must be different, input_file: {input_file_path}, output_file: {output_file_path}"
        )
    if output_file_path is not None and output_file_path != "-":
        if os.path.exists(output_file_path) and not overwrite:
            raise GeodenseError(f"output_file {output_file_path} already exists")
        elif os.path.exists(output_file_path) and overwrite:
            os.remove(output_file_path)


def check_density_file(  # noqa: PLR0913
    input_file_path: str,
    max_segment_length: float,
    density_check_report_path: str | None = None,
    src_crs: str | None = None,
    in_projection: bool = False,
    overwrite: bool = False,
) -> tuple[bool, str, int]:
    if density_check_report_path is None:
        density_check_report_path = os.path.join(
            tempfile.mkdtemp(), "check-density-report.json"
        )

    _validate_dependent_file_args(input_file_path, density_check_report_path, overwrite)

    with open(input_file_path) if input_file_path != "-" else sys.stdin as src:
        geojson_obj = get_geojson_obj(src)
        _geom_type_check(geojson_obj, "check-density")
        has_3d_coords: Has3D = _has_3d_coordinates(geojson_obj)
        geojson_src_crs = _get_crs_geojson(
            geojson_obj, input_file_path, src_crs, has_3d_coords
        )
        config = DenseConfig(
            CRS.from_authority(*geojson_src_crs.split(":")),
            max_segment_length,
            in_projection=in_projection,
        )
        report_fc = density_check_geojson_object(geojson_obj, config)

    failed_segment_count = len(report_fc.features)
    check_status = failed_segment_count == 0

    if not check_status:
        with open(density_check_report_path, "w") as f:
            f.write(report_fc.model_dump_json(indent=4, exclude_none=True))
    return (check_status, density_check_report_path, len(report_fc.features))


def report_line_string_to_geojson(
    report: list[ReportLineString], src_crs_auth_code: str | None
) -> CrsFeatureCollection:
    features = [
        Feature(
            type="Feature",
            properties={"segment_length": x[0]},
            geometry=LineString(type="LineString", coordinates=list(x[1])),
        )
        for x in report
    ]
    result = CrsFeatureCollection(
        features=features, type="FeatureCollection", name="density-check-report"
    )
    if src_crs_auth_code is not None:
        result.set_crs_auth_code(src_crs_auth_code)
    return result


def densify_file(  # noqa: PLR0913
    input_file_path: str,
    output_file_path: str,
    overwrite: bool = False,
    max_segment_length: float | None = None,
    densify_in_projection: bool = False,
    src_crs: str | None = None,
) -> None:
    """_summary_

    Arguments:
        input_file_path: assumes file exists otherwise raises FileNotFoundError
        output_file_path: assumes directory of output_file_path exists

    Keyword Arguments:
        layer -- layer name, when no specified and multilayer file, first layer will be used (default: {None})
        max_segment_length -- max segment length to use for densification (default: {None})
        densify_in_projection -- user src projection for densification (default: {False})
        src_crs -- override src crs of input file (default: {None})

    Raises:
        ValueError: application errors
        pyproj.exceptions.CRSError: when crs cannot be found by pyproj

    """
    _validate_dependent_file_args(input_file_path, output_file_path, overwrite)
    src: TextIO
    with open(input_file_path) if input_file_path != "-" else sys.stdin as src:
        geojson_obj = get_geojson_obj(src)
        has_3d_coords: Has3D = _has_3d_coordinates(geojson_obj)
        geojson_src_crs = _get_crs_geojson(
            geojson_obj, input_file_path, src_crs, has_3d_coords
        )
        config = DenseConfig(
            CRS.from_authority(*geojson_src_crs.split(":")),
            max_segment_length,
            densify_in_projection,
        )
        densify_geojson_object(geojson_obj, config)
        if src_crs is not None and isinstance(geojson_obj, CrsFeatureCollection):
            geojson_obj.set_crs_auth_code(src_crs)
        with open(
            output_file_path, "w"
        ) if output_file_path != "-" else sys.stdout as out_f:
            geojson_obj_model: BaseModel = cast(BaseModel, geojson_obj)

            out_f.write(geojson_obj_model.model_dump_json(indent=1, exclude_none=True))


def apply_function_on_geojson_geometries(  # noqa: C901
    body: Feature
    | CrsFeatureCollection
    | GeojsonGeomNoGeomCollection
    | GeometryCollection,
    callback: Callable[[GeojsonGeomNoGeomCollection], Nested | None],
) -> Nested:
    result: Nested = []

    if isinstance(body, Feature):
        feature = cast(Feature, body)
        if isinstance(feature.geometry, GeometryCollection):
            return apply_function_on_geojson_geometries(feature.geometry, callback)
        geom = cast(GeojsonGeomNoGeomCollection, feature.geometry)
        return [callback(geom)]
    elif isinstance(
        body, Point | MultiPoint | LineString | MultiLineString | Polygon | MultiPolygon
    ):
        geom = cast(GeojsonGeomNoGeomCollection, body)
        return [callback(geom)]
    elif isinstance(body, CrsFeatureCollection):
        fc_body: CrsFeatureCollection = body
        features: Iterable[Feature] = fc_body.features

        for ft in features:
            if ft.geometry is None:
                raise GeodenseError(f"feature does not have a geometry, feature: {ft}")

            result_list = list(result)
            if isinstance(ft.geometry, GeometryCollection):
                ft_result: Nested = apply_function_on_geojson_geometries(
                    ft.geometry, callback
                )
                result_list.append(ft_result)
            else:
                result_list.append(callback(ft.geometry))
            result = result_list
        return result
    elif isinstance(body, GeometryCollection):
        gc = cast(GeometryCollection, body)
        geometries: list[Geometry] = gc.geometries

        for g in geometries:
            if isinstance(g, GeometryCollection):
                raise GeodenseError("nested GeometryCollections are not supported")
            g_no_gc = cast(
                GeojsonGeomNoGeomCollection, g
            )  # geojson prohibits nested geometrycollections - maybe throw exception if this occurs
            result_list = list(result)
            result_list.append(callback(g_no_gc))
            result = result_list
        return result
    return result


def _interpolate_geodesic(
    a: Position, b: Position, densify_config: DenseConfig
) -> LineStringCoords:
    """geodesic interpolate intermediate points between points a and b, with segment_length < max_segment_length. Only returns intermediate points."""

    three_dimensional_points = (
        len(a) == THREE_DIMENSIONAL and len(b) == THREE_DIMENSIONAL
    )
    a_2d: tuple[float, float] = tuple(a[:2])  # type: ignore
    b_2d: tuple[float, float] = tuple(b[:2])  # type: ignore

    transformer = densify_config.transformer

    if (
        densify_config.src_crs.is_projected
    ):  # only convert to basegeographic crs if src_proj is projected
        if transformer is None:
            raise GeodenseError(
                "transformer cannot be None when src_crs.is_projected=True"
            )
        # technically the following transform call is a converion and not a transformation, since crs->base-crs will be a conversion in most cases
        a_t: tuple[float, float] = transformer.transform(*a_2d)
        b_t: tuple[float, float] = transformer.transform(*b_2d)
    else:  # src_crs is geographic do not transform
        a_t, b_t = (a_2d, b_2d)

    g = densify_config.geod

    az12, _, geod_dist = g.inv(*a_t, *b_t, return_back_azimuth=True)
    if math.isnan(geod_dist):
        raise GeodenseError(
            f"unable to calculate geodesic distance, output calculation geodesic distance: {geod_dist}, expected: floating-point number"
        )

    if geod_dist <= densify_config.max_segment_length:
        return []
    else:
        (
            nr_points,
            new_max_segment_length,
        ) = _get_intermediate_nr_points_and_segment_length(
            geod_dist, densify_config.max_segment_length
        )
        r = g.fwd_intermediate(
            *a_t,
            az12,
            npts=nr_points,
            del_s=new_max_segment_length,
            return_back_azimuth=True,
        )

        def optional_back_transform(lon: float, lat: float) -> tuple[Any, Any]:
            """technically should be named optional_back_convert, since crs->base crs is (mostly) a conversion and not a transformation"""
            if densify_config.src_crs.is_projected:
                if densify_config.back_transformer is None:
                    raise GeodenseError(
                        "back_transformer cannot be None when src_crs.is_projected=True"
                    )
                return densify_config.back_transformer.transform(lon, lat)
            return (lon, lat)

        if three_dimensional_points:
            # interpolate height for three_dimensional_points
            a_3d: tuple[float, float, float] = cast(tuple[float, float, float], a)
            b_3d: tuple[float, float, float] = cast(tuple[float, float, float], b)
            height_a = a_3d[2:][0]
            height_b = b_3d[2:][0]
            delta_height_b_a = height_b - height_a
            delta_height_per_point = delta_height_b_a * (
                new_max_segment_length / geod_dist
            )
            return [
                tuple(  # type: ignore
                    (
                        *optional_back_transform(lon, lat),
                        round(
                            (height_a + ((i + 1) * delta_height_per_point)),
                            DEFAULT_PRECISION_METERS,
                        ),
                    )
                )
                for i, (lon, lat) in enumerate(zip(r.lons, r.lats, strict=True))
            ]
        else:
            return [
                optional_back_transform(lon, lat)
                for lon, lat in zip(r.lons, r.lats, strict=True)
            ]


def _get_cartesian_distance(a: Position, b: Position) -> float:
    return math.sqrt((b[0] - a[0]) ** 2 + (b[1] - a[1]) ** 2)  # pythagoras


def _interpolate_src_proj(
    a: Position, b: Position, densify_config: DenseConfig
) -> LineStringCoords:
    """Interpolate intermediate points between points a and b, with segment_length < max_segment_length. Only returns intermediate points."""

    three_dimensional_points = (
        len(a) == THREE_DIMENSIONAL and len(b) == THREE_DIMENSIONAL
    )
    if (
        not three_dimensional_points
    ):  # if not both three dimensional points, ensure both points are two dimensional
        a = a[:2]
        b = b[:2]

    dist = _get_cartesian_distance(a, b)
    if dist <= densify_config.max_segment_length:
        return []
    else:
        new_points: list[tuple[float, float] | tuple[float, float, float]] = []

        (
            nr_points,
            new_max_segment_length,
        ) = _get_intermediate_nr_points_and_segment_length(
            dist, densify_config.max_segment_length
        )

        for i in range(0, nr_points):
            p_point: ShpPoint = ShpLineString([a, b]).interpolate(
                new_max_segment_length * (i + 1)
            )
            p = tuple(p_point.coords[0])
            new_points.append(p)  # type: ignore
        return [
            *new_points,
        ]


def get_geojson_obj(src: TextIO) -> GeojsonObject:
    src_json = json.loads(src.read())
    type_map = {
        "Feature": Feature,
        "GeometryCollection": GeometryCollection,
        "FeatureCollection": CrsFeatureCollection,
        "Point": Point,
        "MultiPoint": MultiPoint,
        "Polygon": Polygon,
        "MultiPolygon": MultiPolygon,
        "LineString": LineString,
        "MultiLineString": MultiLineString,
    }
    try:
        geojson_type = src_json["type"]
        constructor = type_map[geojson_type]
    except KeyError as e:
        message = f'received invalid GeoJSON file, loc: `.type`, value: `{src_json["type"]}`, expected one of: {", ".join(list(type_map.keys()))}'
        raise GeodenseError(message) from e
    geojson_obj: GeojsonObject = constructor(**src_json)
    return geojson_obj


def check_density_geometry_coordinates(
    geometry_coordinates: GeojsonCoordinates,
    densify_config: DenseConfig,
    # indices: list[int] | None = None,
) -> Nested[ReportLineString]:
    # if indices is None:
    #     indices = []

    if isinstance(geometry_coordinates, tuple):
        return [None]

    if _is_linestring_geom(geometry_coordinates):
        linestring_coords = cast(LineStringCoords, geometry_coordinates)
        linestring_report = _check_density_linestring(linestring_coords, densify_config)
        return linestring_report
    else:
        return [
            check_density_geometry_coordinates(e, densify_config)
            for i, e in enumerate(geometry_coordinates)
        ]


class Has3D(Enum):
    all: Literal["all"] = "all"
    some: Literal["some"] = "some"
    none: Literal["none"] = "none"


def _get_crs_geojson(
    geojson_object: GeojsonObject,
    input_file_path: str,
    src_crs: str | None,
    has_3d_coords: Has3D,
) -> str:
    result_crs: str | None = None

    is_fc = False
    if isinstance(geojson_object, CrsFeatureCollection):
        result_crs = geojson_object.get_crs_auth_code()
        is_fc = True

    # case to set default CRS if src_crs not specified and not available in GeoJSON
    if (
        result_crs is None and src_crs is None
    ):  # set default crs if not in geojson object and not overridden with src_crs
        default_crs = DEFAULT_CRS_2D
        if has_3d_coords in [Has3D.some, Has3D.all]:
            default_crs = DEFAULT_CRS_3D
        message = f"unable to determine source CRS for file {input_file_path}, assumed CRS is {default_crs}"
        logger.warning(message)
        result_crs = default_crs
        if is_fc:
            fc: CrsFeatureCollection = cast(CrsFeatureCollection, geojson_object)
            fc.set_crs_auth_code(result_crs)

    # is src_crs is set use src_crs
    elif src_crs is not None:
        src_crs_crs: CRS = CRS.from_authority(*src_crs.split(":"))
        if has_3d_coords == Has3D.all and not src_crs_crs.is_vertical:
            logger.warning(
                "src_crs is 2D while input data contains geometries with 3D coordinates"
            )
        result_crs = (
            src_crs if src_crs is not None else result_crs
        )  # override json_crs with src_crs if defined

    elif result_crs is None:
        raise GeodenseError("could not determine CRS from GeoJSON object")

    return result_crs


def flatten(container: Nested) -> Iterable:
    for i in container:
        if isinstance(i, Sequence) and not isinstance(i, tuple):
            yield from flatten(i)
        else:
            yield i


def _check_density_linestring(
    linestring: LineStringCoords,
    densify_config: DenseConfig,
    # indices: list[int],
) -> list[ReportLineString]:
    result = []

    for k in range(0, len(linestring) - 1):
        a: Position = linestring[k]
        b: Position = linestring[k + 1]

        a_2d: tuple[float, float] = tuple(a[0:2])  # type: ignore
        b_2d: tuple[float, float] = tuple(b[0:2])  # type: ignore

        if densify_config.in_projection:
            linesegment_dist = _get_cartesian_distance(a_2d, b_2d)
        else:
            if (
                densify_config.src_crs.is_projected
            ):  # only convert to basegeographic crs if src_proj is projected
                transformer = densify_config.transformer
                if transformer is None:
                    raise GeodenseError(
                        "transformer cannot be None when src_crs.is_projected=True"
                    )
                a_t = transformer.transform(*a_2d)
                b_t = transformer.transform(*b_2d)
            else:  # src_crs is geographic do not transform
                a_t, b_t = (a_2d, b_2d)
            g = densify_config.geod

            # see https://github.com/python/mypy/issues/6799
            _, _, geod_dist = g.inv(*a_t, *b_t, return_back_azimuth=True)  # type: ignore
            if math.isnan(geod_dist):
                raise GeodenseError(
                    f"unable to calculate geodesic distance, output calculation geodesic distance: {geod_dist}, expected: floating-point number"
                )
            linesegment_dist = geod_dist
        if linesegment_dist > (densify_config.max_segment_length + 0.001):
            result.append((linesegment_dist, (a, b)))
    return result


def _is_linestring_geom(geometry_coordinates: GeojsonCoordinates) -> bool:
    """Check if coordinates are of linestring geometry type.

        - Fiona linestring coordinates are of type: list[tuple[float,float,...]])
        - GeoJSON linestring coordinates are of type: list[list[float]]

    Args:
        geometry_coordinates (list): Fiona or GeoJSON coordinates sequence

    Returns:
        bool: if geometry_coordinates is linestring return True else False
    """
    if (
        len(geometry_coordinates) > 0
        and isinstance(geometry_coordinates[0], Sequence)
        and all(
            isinstance(x, float | int) for x in geometry_coordinates[0]
        )  # also test for int just in case...
    ):
        return True
    return False


def _raise_e_if_point_geom(geometry_coordinates: GeojsonCoordinates) -> None:
    if isinstance(geometry_coordinates, tuple):  # assume is point when tuple
        raise GeodenseError(
            "received point geometry coordinates, instead of (multi)linestring"
        )


def _transform_linestrings_in_geometry_coordinates(
    geometry_coordinates: GeojsonCoordinates,
    transform_fun: Callable[[LineStringCoords], Nested],
    retain_point_geoms: bool = False,
) -> Nested | None | T:
    if isinstance(geometry_coordinates, tuple):
        if retain_point_geoms:
            return geometry_coordinates
        else:
            return None
    if _is_linestring_geom(geometry_coordinates):
        linestring_coords = cast(LineStringCoords, geometry_coordinates)
        return transform_fun(linestring_coords)
    else:
        return [
            _transform_linestrings_in_geometry_coordinates(
                e, transform_fun, retain_point_geoms=retain_point_geoms
            )
            for e in geometry_coordinates
        ]


def _get_intermediate_nr_points_and_segment_length(
    dist: float, max_segment_length: float
) -> tuple[int, float]:
    if dist <= max_segment_length:
        raise GeodenseError(
            f"max_segment_length ({max_segment_length}) cannot be bigger or equal than dist ({dist})"
        )
    remainder = dist % max_segment_length
    nr_segments = int(dist // max_segment_length)
    if remainder > 0:
        nr_segments += 1
    new_max_segment_length = dist / nr_segments  # space segments evenly over delta(a,b)
    nr_points = (
        nr_segments - 1
    )  # convert nr of segments to nr of intermediate points, should be at least 1
    return nr_points, new_max_segment_length


def _add_vertices_to_line_segment(
    linestring: LineStringCoords, coord_index: int, densify_config: DenseConfig
) -> int:
    """Adds vertices to linestring in place, and returns number of vertices added to linestring.

    Args:
        ft_linesegment (_type_): line segment to add vertices
        coord_index (int): coordinate index of line segment to add vertices for
        transformer (Transformer): pyproj transformer
        max_segment_length (float): max segment length, if exceeded vertices will be added
        densify_in_projection (bool): whether to use source projection to densify (not use great-circle distance)

    Returns:
        int: number of added vertices
    """

    a = linestring[coord_index]
    b = linestring[coord_index + 1]

    prec = densify_config.get_coord_precision()

    if not densify_config.in_projection:
        p = list(
            [
                _round_coordinates(x, prec)
                for x in _interpolate_geodesic(a, b, densify_config)
            ]
        )
    else:
        p = list(
            [
                _round_coordinates(x, prec)
                for x in _interpolate_src_proj(a, b, densify_config)
            ]
        )

    linestring[coord_index] = _round_coordinates(linestring[coord_index], prec)  # type: ignore
    linestring[coord_index + 1] = _round_coordinates(linestring[coord_index + 1], prec)  # type: ignore
    linestring[coord_index + 1 : coord_index + 1] = p  # type: ignore
    return len(p)


def _round_coordinates(coordinates: tuple, position_precision: int) -> tuple:
    result = tuple([round(x, position_precision) for x in coordinates[:2]])
    if len(coordinates) == THREE_DIMENSIONAL:
        result = (*result, round(coordinates[2], DEFAULT_PRECISION_METERS))
    return result


def _get_geometry_type(
    geometry: GeojsonGeomNoGeomCollection,
    _indices: list[int] | None = None,
) -> str:
    return cast(str, geometry.type)


def _geom_has_3d_coords(
    geometry: GeojsonGeomNoGeomCollection,
) -> Nested[bool] | None:
    def _linestring_has_3d_coords(linestring_coords: LineStringCoords) -> Nested[bool]:
        return [len(x) == THREE_DIMENSIONAL for x in linestring_coords]

    return _transform_linestrings_in_geometry_coordinates(
        geometry.coordinates, _linestring_has_3d_coords
    )


def _get_line_segment_densify_fun(
    densify_config: DenseConfig,
) -> Callable[[LineStringCoords], LineStringCoords]:
    def line_segment_densify(
        linestring: LineStringCoords,
    ) -> LineStringCoords:
        added_nodes = 0
        stop = len(linestring) - 1
        for i, _ in enumerate(linestring[:stop]):
            added_nodes += _add_vertices_to_line_segment(
                linestring, i + added_nodes, densify_config
            )
        return linestring

    return line_segment_densify


def _get_geom_densify_fun(
    densify_config: DenseConfig,
) -> Callable[[GeojsonGeomNoGeomCollection], GeojsonCoordinates]:
    def _geom_densify(geometry: GeojsonGeomNoGeomCollection) -> GeojsonCoordinates:
        _add_vertices_exceeding_max_segment_length = _get_line_segment_densify_fun(
            densify_config
        )
        result: GeojsonCoordinates = _transform_linestrings_in_geometry_coordinates(  # type: ignore
            geometry.coordinates,
            _add_vertices_exceeding_max_segment_length,
            retain_point_geoms=True,
        )
        geometry.coordinates = result
        return result

    return _geom_densify


def _has_3d_coordinates(
    geojson_obj: GeojsonObject, silent: bool | None = False
) -> Has3D:
    has_3d_coords: Nested[bool] = apply_function_on_geojson_geometries(
        geojson_obj, _geom_has_3d_coords
    )
    has_3d_coords_flat = list(flatten(has_3d_coords))
    if not all(has_3d_coords_flat) and any(has_3d_coords_flat):  # some 3d
        if not silent:
            warning_message = "geometries with mixed 2D and 3D vertices found"
            logger.warning(warning_message)
        return Has3D.some
    elif all(not x for x in has_3d_coords_flat):  # none 3d
        return Has3D.none
    return Has3D.all


def _geom_type_check(geojson_obj: GeojsonObject, command: str = "") -> None:
    geom_types: Nested[str] = apply_function_on_geojson_geometries(
        geojson_obj, _get_geometry_type
    )

    if all(g_t in ("Point", "MultiPoint") for g_t in geom_types):
        # situation: all geoms point -> error
        if command:
            error_message = f"cannot run {command} on GeoJSON that only contains (Multi)Point geometries"
        else:
            error_message = "GeoJSON contains only (Multi)Point geometries"
        raise GeodenseError(error_message)
    elif any(gt in ["Point", "MultiPoint"] for gt in geom_types):
        # sitation: some geoms point -> warning
        warning_message = "GeoJSON contains (Multi)Point geometries"
        if command:
            warning_message = (
                f"{warning_message}, cannot run {command} on (Multi)Point geometries"
            )

        logger.warning(warning_message)
    else:
        # situation: no geoms point -> ok
        pass



import logging
import os

import pytest
from geodense.lib import get_geojson_obj


@pytest.fixture()
def test_dir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture()
def linestring_d10_feature_gj(test_dir):
    with open(os.path.join(test_dir, "data", "linestring_feature_d10.json")) as f:
        return get_geojson_obj(f)


@pytest.fixture()
def linestring_feature_gj(test_dir):
    with open(os.path.join(test_dir, "data", "linestring_feature.json")) as f:
        return get_geojson_obj(f)


@pytest.fixture()
def linestring_3d_feature_gj(test_dir):
    with open(os.path.join(test_dir, "data", "linestring_3d_feature.json")) as f:
        return get_geojson_obj(f)


@pytest.fixture()
def linestring_feature_5000_gj(test_dir):
    with open(os.path.join(test_dir, "data", "linestring_feature_5000.json")) as f:
        return get_geojson_obj(f)


@pytest.fixture()
def geometry_collection_gj(test_dir):
    with open(os.path.join(test_dir, "data", "feature-geometry-collection.json")) as f:
        return get_geojson_obj(f)


@pytest.fixture()
def polygon_feature_with_holes_gj(test_dir):
    with open(os.path.join(test_dir, "data", "polygon_feature_with_holes.json")) as f:
        return get_geojson_obj(f)


@pytest.fixture()
def point_feature_gj(test_dir):
    with open(os.path.join(test_dir, "data", "point_feature.json")) as f:
        return get_geojson_obj(f)


@pytest.fixture()
def geometry_collection_feature_gj(test_dir):
    with open(os.path.join(test_dir, "data", "feature-geometry-collection.json")) as f:
        return get_geojson_obj(f)


@pytest.fixture()
def linestring_feature_multiple_linesegments(test_dir):
    with open(
        os.path.join(test_dir, "data", "linestring_feature_multiple_linesegments.json")
    ) as f:
        return get_geojson_obj(f)


geodense_logger = logging.getLogger("geodense")


# see https://github.com/pytest-dev/pytest/issues/14#issuecomment-521577819
# required since: "I think the issue boils down to the fact that if you set up a Stream Handler and you enable capsys the actual output stream will be a pytest stream that will be closed and thrown away at the next test in the test suite."
@pytest.fixture(autouse=True)
def _ensure_logging_framework_not_altered():
    before_handlers = list(geodense_logger.handlers)
    yield
    geodense_logger.handlers = before_handlers



import pytest
from geodense.lib import (
    _add_vertices_to_line_segment,
    _get_line_segment_densify_fun,
    _interpolate_src_proj,
)
from geodense.models import (
    DEFAULT_PRECISION_DEGREES,
    DEFAULT_PRECISION_METERS,
    DenseConfig,
)
from pyproj import CRS


def test_interpolate_src_proj_no_op():
    points = [(0, 0), (10, 10)]  # 14.142
    c = DenseConfig(CRS.from_epsg(28992), 20)
    points_t = _interpolate_src_proj(*points, c)
    assert points_t == [], f"expected points_t to be empty, received: {points_t}"


def test_interpolate_src_proj():
    points = [(0, 0), (10, 10)]  # 14.142
    c = DenseConfig(CRS.from_epsg(28992), 10)
    points_t = _interpolate_src_proj(*points, c)
    expected_nr_of_points = 1

    assert (
        len(points_t) == expected_nr_of_points
    ), f"expected length of points_t was {expected_nr_of_points}, acual: {len(points_t)}"


def test_add_vertices_exceeding_max_segment_length():
    linestring = [(0, 0), (10, 10), (20, 20)]
    linestring_t = linestring

    c = DenseConfig(CRS.from_epsg(28992), 10, True)
    line_segment_densify = _get_line_segment_densify_fun(c)
    line_segment_densify(linestring_t)
    assert len(linestring_t) == 5  # noqa: PLR2004
    assert linestring_t == [(0, 0), (5.0, 5.0), (10, 10), (15.0, 15.0), (20, 20)]


def test_interpolate_round_projected():
    """Note precision is only reduced by round()"""
    points_proj = [(0.12345678, 0.12345678), (10.12345678, 10.12345678)]
    c = DenseConfig(CRS.from_epsg(28992), 10, True)
    _add_vertices_to_line_segment(points_proj, 0, c)

    assert all(
        [
            str(x)[::-1].find(".") == DEFAULT_PRECISION_METERS
            for p in points_proj
            for x in p
        ]
    )  # https://stackoverflow.com/a/26231848/1763690


def test_interpolate_round_geographic():
    """Note precision is only reduced by round()"""
    points_geog = [
        (0.1234567891011, 0.1234567891011),
        (10.1234567891011, 10.1234567891011),
    ]
    c = DenseConfig(CRS.from_epsg(4258), 10)
    _add_vertices_to_line_segment(points_geog, 0, c)

    assert all(
        [
            str(x)[::-1].find(".")
            <= DEFAULT_PRECISION_DEGREES  # <= because Python does not force n precision, see for instance: `round(2.000000,9) -> 2.0`
            for p in points_geog
            for x in p
        ]
    )  # https://stackoverflow.com/a/26


@pytest.mark.parametrize(
    ("linestring", "in_proj", "expectation"),
    [
        (
            [(-10, -10), (0, 0, 0), (10, 10, 10), (20, 20), (30, 30)],
            False,
            [
                (-10, -10),
                (-5.0, -5.0),
                (0, 0, 0),
                (5.0, 5.0, 5.0),
                (10, 10, 10),
                (15.0, 15.0),
                (20, 20),
                (25.0, 25.0),
                (30, 30),
            ],
        ),
        (
            [(-10, -10), (0, 0, 0), (10, 10, 10), (20, 20), (30, 30)],
            True,
            [
                (-10, -10),
                (-5.0, -5.0),
                (0, 0, 0),
                (5.0, 5.0, 5.0),
                (10, 10, 10),
                (15.0, 15.0),
                (20, 20),
                (25.0, 25.0),
                (30, 30),
            ],
        ),
    ],
)
def test_interpolate_3d(linestring, in_proj, expectation):
    linestring_t = linestring[:]  # clone list with slicing
    c = DenseConfig(CRS.from_epsg(7415), 10, in_proj)

    line_segment_densify = _get_line_segment_densify_fun(c)
    line_segment_densify(linestring_t)
    assert linestring_t == expectation



import contextlib
import logging
import sys
from importlib.metadata import PackageNotFoundError, version
from logging import Formatter, NullHandler, StreamHandler

logging.getLogger(__name__).addHandler(NullHandler())

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("geodense")


def get_log_handler(verbose: bool) -> logging.StreamHandler:
    formatter = get_formatter(verbose)
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setFormatter(formatter)
    return handler


def get_formatter(verbose: bool) -> Formatter:
    format_string_default = "[%(levelname)s] %(message)s"
    format_string_verbose = f"[%(levelname)s] [%(filename)s:%(lineno)d] [{__name__}.%(module)s.%(funcName)s] %(message)s"
    format_string = format_string_verbose if verbose else format_string_default
    formatter = logging.Formatter(format_string)
    return formatter


# copied from https://github.com/urllib3/urllib3/blob/f7cd7f3f632cf5224f627536f02c2abf7e4146d1/src/urllib3/__init__.py and adapted
# error handling convention:
# expected caught errors -> log as error
# unexpected caugh errors (this is the case for a generic error handler) -> log as exception
def add_stderr_logger(verbose: bool = False) -> StreamHandler:
    """
    Adding streamhandler to the logger, used for CLI output to stderr.

    Returns the handler after adding it.
    """
    # This method needs to be in this __init__.py to get the __name__ correct
    # even if geodense is vendored within another package.
    # verbose: False
    #   - loglevel:  WARNING
    #   - format_string: default
    # verbose: True
    #   - loglevel:  DEBUG
    #   - format_string: verbose
    level = logging.DEBUG if verbose else logging.WARNING
    handler = get_log_handler(verbose)
    logger = logging.getLogger(__name__)
    logger.addHandler(handler)
    logger.setLevel(level)
    return handler



from pyproj import CRS as ProjCrs  # noqa: N811
from pyproj import Transformer

DEFAULT_MAX_SEGMENT_LENGTH = 200
DEFAULT_PRECISION_DEGREES = 9
DEFAULT_PRECISION_METERS = 4


class GeodenseError(Exception):
    pass


class DenseConfig:
    geod_empty_exc_message = "DenseConfig.geod is None, cannot perform geodetic calculations without pyproj.Geod object"

    def __init__(
        self: "DenseConfig",
        src_crs: ProjCrs,
        max_segment_length: float | None = None,
        in_projection: bool = False,
    ) -> None:
        self.src_crs = src_crs

        if self.src_crs.is_geographic and in_projection:
            raise GeodenseError(
                f"in_projection can only be used with \
projected coordinates reference systems, crs {self.src_crs} is a geographic crs"
            )
        self.back_transformer = None
        if self.src_crs.is_geographic:
            _geod = self.src_crs.get_geod()
            self.transformer = None
        elif self.src_crs.is_projected:
            base_crs = self._get_base_crs()
            self.transformer = Transformer.from_crs(src_crs, base_crs, always_xy=True)
            _geod = self.src_crs.get_geod()
            self.back_transformer = Transformer.from_crs(
                base_crs, src_crs, always_xy=True
            )
        else:
            raise GeodenseError(
                "unexpected crs encountered, crs is neither geographic nor projected"
            )
        if _geod is None:
            raise GeodenseError(self.geod_empty_exc_message)
        self.geod = _geod

        self.in_projection = in_projection
        self.max_segment_length = abs(
            max_segment_length or DEFAULT_MAX_SEGMENT_LENGTH
        )  # when max_segment_length == None -> DEFAULT_MAX_SEGMENT_LENGTH

    def _get_base_crs(self: "DenseConfig") -> ProjCrs:
        if self.src_crs is None:
            raise GeodenseError("field src_proj is None")

        if self.src_crs.is_geographic:
            raise GeodenseError("cannot get base_crs for geographic crs")

        crs_dict = self.src_crs.to_json_dict()
        if crs_dict["type"] == "ProjectedCRS":
            base_crs_id = self.src_crs.to_json_dict()["base_crs"]["id"]
        elif crs_dict["type"] == "CompoundCRS":
            projected_crs = next(
                x for x in crs_dict["components"] if x["type"] == "ProjectedCRS"
            )
            base_crs_id = projected_crs["base_crs"]["id"]
        return ProjCrs.from_authority(base_crs_id["authority"], base_crs_id["code"])

    def get_coord_precision(self: "DenseConfig") -> int:
        if self.src_crs is None:
            raise GeodenseError("DensifyConfig.source_crs is None")
        return (
            DEFAULT_PRECISION_DEGREES
            if self.src_crs.is_geographic
            else DEFAULT_PRECISION_METERS
        )



import re
from typing import Literal

from geojson_pydantic import FeatureCollection
from pydantic import BaseModel, Field

from geodense.models import GeodenseError


class GeoJsonCrsProp(BaseModel):
    # OGC URN scheme - 8.2 in OGC 05-103
    # urn:ogc:def:crs:{crs_auth}:{crs_version}:{crs_identifier}
    name: str = Field(pattern=r"^urn:ogc:def:crs:.*?:.*?:.*?$")


class GeoJsonCrs(BaseModel):
    properties: GeoJsonCrsProp
    type: Literal["name"] = "name"


class CrsFeatureCollection(FeatureCollection):
    crs: GeoJsonCrs | None = None
    name: str | None = None

    def set_crs_auth_code(self: "CrsFeatureCollection", crs_auth_code: str) -> None:
        crs_auth, crs_identifier = crs_auth_code.split(":")
        geojson_crs_val = f"urn:ogc:def:crs:{crs_auth}::{crs_identifier}"
        if self.crs is None:
            self.crs = GeoJsonCrs(properties=GeoJsonCrsProp(name=geojson_crs_val))
        self.crs.properties.name = geojson_crs_val

    def get_crs_auth_code(self: "CrsFeatureCollection") -> str | None:
        if self.crs is None:
            return None
        result = re.search(
            r"^urn:ogc:def:crs:(.*?):.*?:(.*?)$", self.crs.properties.name
        )
        if result is None:
            raise GeodenseError(
                f"unable to retrieve crs_auth_code from crs.properties.name {self.crs.properties.name}"
            )
        return f"{result.group(1)}:{result.group(2)}"



import re
import tempfile

import pytest
from cli_test_helpers import ArgvContext
from geodense.main import main


# @pytest.mark.usefixtures("clear_log") # see https://github.com/pytest-dev/pytest/issues/5502#issuecomment-1509988433
def test_3d_coords_2d_crs_raises_warning(test_dir, capsys):
    in_file = "linestring_3d_feature.json"
    in_filepath = f"{test_dir}/data/{in_file}"
    max_segment_length = "5000"
    with ArgvContext(
        "geodense",
        "check-density",
        in_filepath,
        "--max-segment-length",
        max_segment_length,
        "--src-crs",
        "EPSG:28992",
    ), pytest.raises(SystemExit):
        main()
    _, err = capsys.readouterr()
    assert (
        err
        == "[WARNING] src_crs is 2D while input data contains geometries with 3D coordinates\n"
    )


@pytest.mark.parametrize(
    ("input_file", "src_crs", "expected"),
    [
        (
            "geometry-4326-no-crs.json",
            None,
            r"\[WARNING\] unable to determine source CRS for file .*, assumed CRS is OGC:CRS84\n",
        ),
        ("linestring_feature.json", "EPSG:7415", r""),
    ],
)
def test_2d_coords_3d_crs_raises_no_warning(
    test_dir, capsys, input_file, src_crs, expected
):
    in_filepath = f"{test_dir}/data/{input_file}"
    max_segment_length = "10000"
    args = [
        "geodense",
        "check-density",
        in_filepath,
        "--max-segment-length",
        max_segment_length,
    ]
    if src_crs is not None:
        args.extend(["--src-crs", src_crs])
    with ArgvContext(*args), pytest.raises(SystemExit):
        main()
    _out, err = capsys.readouterr()
    assert re.match(expected, err)


@pytest.mark.parametrize(
    ("input_file", "src_crs", "expected"),
    [
        (
            "geometry-4326-no-crs.json",
            None,
            r"\[WARNING\] unable to determine source CRS for file .*, assumed CRS is OGC:CRS84\n",
        ),
        ("linestring_feature.json", "EPSG:7415", r""),
    ],
)
def test_2d_coords_3d_crs_raises_no_warning_densify(
    test_dir, capsys, input_file, src_crs, expected
):
    in_filepath = f"{test_dir}/data/{input_file}"
    max_segment_length = "5000"
    out_dir = tempfile.mkdtemp()
    out_filepath = f"{out_dir}/{input_file}"
    args = [
        "geodense",
        "densify",
        "--max-segment-length",
        max_segment_length,
        in_filepath,
        out_filepath,
    ]
    if src_crs is not None:
        args.extend(["--src-crs", src_crs])
    with ArgvContext(*args):
        main()
    _out, err = capsys.readouterr()
    assert re.match(expected, err)



import json
import os
import re
import tempfile
from contextlib import nullcontext as does_not_raise
from contextlib import suppress

import pyproj
import pytest
from geodense.lib import (
    _get_geom_densify_fun,
    _get_intermediate_nr_points_and_segment_length,
    densify_file,
    densify_geojson_object,
    get_geojson_obj,
)
from geodense.models import DenseConfig, GeodenseError
from geojson_pydantic import Feature, GeometryCollection


@pytest.mark.parametrize(
    ("geojson", "expected"),
    [
        (
            "point_feature_gj",
            pytest.raises(
                GeodenseError,
                match=r"cannot run densify on GeoJSON that only contains \(Multi\)Point geometries",
            ),
        ),
        (
            "geometry_collection_gj",
            does_not_raise(),
        ),  # geom collection containing point
    ],
)
def test_only_points_raises_exception(geojson, expected, request):
    feature: Feature = request.getfixturevalue(geojson)
    c = DenseConfig(pyproj.CRS.from_epsg(28992))

    with expected:
        densify_geojson_object(feature, c)


def test_geometry_collection_transformed(geometry_collection_gj):
    gc: GeometryCollection = geometry_collection_gj
    gc_t: GeometryCollection = gc.model_copy(deep=True)

    c = DenseConfig(pyproj.CRS.from_epsg(28992), 10)
    densify_geojson_object(gc_t, c)

    assert gc_t != gc


def test_linestring_d10_transformed(linestring_d10_feature_gj):
    ft: Feature = linestring_d10_feature_gj
    ft_t: Feature = ft.model_copy(deep=True)

    c = DenseConfig(pyproj.CRS.from_epsg(28992), 10)
    geom_densify_fun = _get_geom_densify_fun(c)
    geom_densify_fun(ft_t.geometry)
    feature_coords_length = 2
    feature_t_coord_length = 3

    assert ft != ft_t
    assert len(ft.geometry.coordinates) == feature_coords_length
    assert len(ft_t.geometry.coordinates) == feature_t_coord_length


def test_linestring_d10_no_op(linestring_d10_feature_gj):
    ft: Feature = linestring_d10_feature_gj
    ft_t: Feature = ft.model_copy(deep=True)
    ft_t_in_proj: Feature = ft.model_copy(deep=True)
    ft_t.properties[
        "foo"
    ] = "bar"  # to differentiate between original ft and ft_t, to prevent we are testing the same object...

    c = DenseConfig(
        pyproj.CRS.from_epsg(28992), 15
    )  # 15 since 15> sqrt(10^2+10^2) -- see linestring_feature_d10.json -> this should result in no points added
    c_in_proj = DenseConfig(pyproj.CRS.from_epsg(28992), 15, True)

    geom_densify_fun = _get_geom_densify_fun(c)
    geom_densify_in_proj_fun = _get_geom_densify_fun(c_in_proj)

    geom_densify_fun(ft_t.geometry)
    geom_densify_in_proj_fun(ft_t_in_proj.geometry)

    feature_coords_length = 2

    assert ft != ft_t
    assert len(ft_t.geometry.coordinates) == feature_coords_length
    assert len(ft_t_in_proj.geometry.coordinates) == feature_coords_length


def test_linestring_transformed(linestring_feature_gj):
    ft: Feature = linestring_feature_gj
    ft_t: Feature = ft.model_copy(deep=True)
    c = DenseConfig(pyproj.CRS.from_epsg(28992), 1000)

    geom_densify_fun = _get_geom_densify_fun(c)
    geom_densify_fun(ft_t.geometry)

    feature_coords_length = 2
    feature_t_coord_length = 12

    assert ft != ft_t
    assert len(ft.geometry.coordinates) == feature_coords_length
    assert len(ft_t.geometry.coordinates) == feature_t_coord_length


def test_densify_3d_source_projection(linestring_3d_feature_gj):
    ft_t: Feature = linestring_3d_feature_gj
    c = DenseConfig(pyproj.CRS.from_epsg(7415), 1000, True)

    geom_densify_fun = _get_geom_densify_fun(c)
    geom_densify_fun(ft_t.geometry)

    for i, (_, _, h) in enumerate(ft_t.geometry.coordinates):
        assert h == (
            i * 10
        ), f"height is not linear interpolated with increments of 10 - height: {h}, expected height: {i*10}"


def test_densify_3d(linestring_3d_feature_gj):
    ft_t: Feature = linestring_3d_feature_gj
    c = DenseConfig(pyproj.CRS.from_epsg(7415), 1000)

    geom_densify_fun = _get_geom_densify_fun(c)
    geom_densify_fun(ft_t.geometry)

    for i, (_, _, h) in enumerate(ft_t.geometry.coordinates):
        assert h == (
            i * 10
        ), f"height is not linear interpolated with increments of 10 - height: {h}, expected height: {i*10}"


def test_polygon_with_hole_transformed(polygon_feature_with_holes_gj):
    feature: Feature = polygon_feature_with_holes_gj
    feature_t = feature.model_copy(deep=True)

    c = DenseConfig(pyproj.CRS.from_epsg(28992), 1000)

    geom_densify_fun = _get_geom_densify_fun(c)
    geom_densify_fun(feature_t.geometry)

    assert feature != feature_t


def test_linestring_transformed_source_proj(linestring_feature_gj):
    feature: Feature = linestring_feature_gj
    feature_t = feature.model_copy(deep=True)

    c = DenseConfig(pyproj.CRS.from_epsg(28992), 1000, True)

    geom_densify_fun = _get_geom_densify_fun(c)
    geom_densify_fun(feature_t.geometry)

    feature_coords_length = 2
    feature_t_coord_length = 12

    assert feature != feature_t
    assert len(feature.geometry.coordinates) == feature_coords_length
    assert len(feature_t.geometry.coordinates) == feature_t_coord_length


def test_maxlinesegment_param(linestring_feature_gj):
    feature_200: Feature = linestring_feature_gj
    feature_1000 = feature_200.model_copy(deep=True)

    c_200 = DenseConfig(pyproj.CRS.from_epsg(28992), 200, False)
    c_1000 = DenseConfig(pyproj.CRS.from_epsg(28992), 1000, False)

    geom_densify_fun_200 = _get_geom_densify_fun(c_200)
    geom_densify_fun_1000 = _get_geom_densify_fun(c_1000)

    # note geom_densify_fun returns coordinates and modifies coords in place
    # maybe change to keep it more fucntional and let the calling code update the coords
    coords_200 = geom_densify_fun_200(feature_200.geometry)
    coords_1000 = geom_densify_fun_1000(feature_1000.geometry)

    assert coords_200 == feature_200.geometry.coordinates
    assert coords_1000 == feature_1000.geometry.coordinates
    assert len(feature_200.geometry.coordinates) > len(
        feature_1000.geometry.coordinates
    )


@pytest.mark.parametrize(
    ("input_args", "expectation"),
    [((1000, 100), (9, 100.0)), ((1000, 300), (3, 250.0))],
)
def test_get_intermediate_nr_points_and_segment_length(input_args, expectation):
    # test distance > max_segment_length
    result = _get_intermediate_nr_points_and_segment_length(*input_args)
    assert result == expectation


@pytest.mark.parametrize(
    ("input_args", "expectation"),
    [
        (
            (100, 1000),
            pytest.raises(
                GeodenseError,
                match=r"max_segment_length \(.+\) cannot be bigger or equal than dist \(.+\)",
            ),
        ),
        (
            (1000, 1000),
            pytest.raises(
                GeodenseError,
                match=r"max_segment_length \(.+\) cannot be bigger or equal than dist \(.+\)",
            ),
        ),
        ((1000, 100), does_not_raise()),
    ],
)
def test_get_intermediate_nr_points_and_segment_length_raises(input_args, expectation):
    with expectation:
        _get_intermediate_nr_points_and_segment_length(*input_args)


def test_densify_file(test_dir):
    in_file = "linestrings.json"
    out_file = os.path.join(tempfile.mkdtemp(), in_file)
    densify_file(os.path.join(test_dir, "data", in_file), out_file)
    assert os.path.exists(out_file)


@pytest.mark.parametrize(
    ("input_file", "src_crs", "expectation"),
    [
        (
            "multipolygon.json",
            None,
            "urn:ogc:def:crs:EPSG::28992",
        ),
        ("geometry-4326-no-crs.json", None, None),
        ("geometry-4326-no-crs.json", "OGC:CRS84", None),
        (
            "multipolygon_wrong_crs.json",
            "EPSG:28992",
            "urn:ogc:def:crs:EPSG::28992",
        ),
    ],
)
def test_densify_file_crs_matches_input(test_dir, input_file, src_crs, expectation):
    output_file = os.path.join(tempfile.mkdtemp(), input_file)
    input_file = os.path.join(test_dir, "data", input_file)
    densify_file(input_file, output_file, src_crs=src_crs)
    with open(output_file) as f:
        geojson = json.load(f)
        crs = None
        with suppress(KeyError):
            crs = geojson["crs"]["properties"]["name"]
        assert crs == expectation


def test_densify_file_negative(tmpdir, test_dir):
    in_file = "linestrings.json"
    out_file = os.path.join(tmpdir, in_file)
    densify_file(os.path.join(test_dir, "data", in_file), out_file, None, -10000)
    assert os.path.exists(out_file)


def test_densify_file_in_proj_exc(tmpdir, test_dir):
    in_file = "linestrings-4258.json"
    out_file = os.path.join(tmpdir, in_file)
    in_file = os.path.join(test_dir, "data", in_file)
    with pytest.raises(
        GeodenseError,
        match=(
            r"in_projection can only be used with projected coordinates "
            r"reference systems, crs .+ is a geographic crs"
        ),
    ):
        densify_file(in_file, out_file, densify_in_projection=True)


def test_densify_file_input_file_does_not_exist(tmpdir, test_dir):
    input_file = os.path.join(test_dir, "foobar.json")
    output_file = os.path.join(tmpdir, "foobar.json")

    with pytest.raises(
        FileNotFoundError,
    ):
        densify_file(input_file, output_file)


def test_densify_file_output_dir_does_not_exist_raises(test_dir):
    input_file = os.path.join(test_dir, "data", "linestrings.json")
    output_file = os.path.join(test_dir, "foobar", "foobar.json")

    with pytest.raises(FileNotFoundError):
        assert isinstance(densify_file(input_file, output_file))


def test_densify_file_input_and_output_equal_raises(test_dir):
    input_file = os.path.join(test_dir, "data", "linestrings.json")
    output_file = os.path.join(test_dir, "data", "linestrings.json")

    with pytest.raises(
        GeodenseError,
        match=(
            r"input_file and output_file arguments must be different, input_file: .*linestrings.json, output_file: .*linestrings.json"
        ),
    ):
        assert isinstance(densify_file(input_file, output_file))


@pytest.mark.parametrize(
    ("overwrite", "expectation"),
    [
        (
            False,
            pytest.raises(
                GeodenseError,
                match=r"output_file .*\.json already exists",
            ),
        ),
        (True, does_not_raise()),
    ],
)
def test_densify_file_output_file_exists_raises(test_dir, overwrite, expectation):
    input_file = os.path.join(test_dir, "data", "linestrings.json")
    with tempfile.NamedTemporaryFile(suffix=".json") as tmp_file, expectation:
        densify_file(input_file, tmp_file.name, overwrite=overwrite)


@pytest.mark.parametrize(
    ("crs", "expectation"),
    [
        (
            "EPSG:999999928992",
            pytest.raises(
                pyproj.exceptions.CRSError,
                match=r"Invalid projection: EPSG:999999928992: \(Internal Proj Error: proj_create: crs not found\)",
            ),
        ),
        ("EPSG:28992", does_not_raise()),
    ],
)
def test_densify_file_invalid_crs_raises(test_dir, tmpdir, crs, expectation):
    input_file = os.path.join(test_dir, "data", "feature-geometry-collection.json")
    output_file = os.path.join(tmpdir, "geometry.json")
    with expectation:
        assert densify_file(input_file, output_file, src_crs=crs) is None


def test_point_raises_warning_and_noop(test_dir, tmpdir, caplog):
    input_file = os.path.join(test_dir, "data", "feature-geometry-collection.json")
    output_file = os.path.join(tmpdir, "geometry.json")
    densify_file(input_file, output_file, src_crs="EPSG:28992")
    output = caplog.text
    expected_warning = r"WARNING .* GeoJSON contains \(Multi\)Point geometries, cannot run densify on \(Multi\)Point geometries.*"
    assert re.match(
        expected_warning, output
    ), f"stderr expected message is: '{expected_warning}', actual message was: '{output}'"

    with open(input_file) as in_f, open(output_file) as out_f:
        ft_gc: Feature = get_geojson_obj(in_f)
        ft_gc_t: Feature = get_geojson_obj(out_f)
        assert (
            ft_gc.geometry.geometries[0].coordinates
            == ft_gc_t.geometry.geometries[0].coordinates
        )  # assert point geom coords the same


def test_densify_file_json_no_crs_outputs_message_stderr(caplog, test_dir, tmpdir):
    input_file = os.path.join(test_dir, "data", "geometry-4326-no-crs.json")
    output_file = os.path.join(tmpdir, "geometry.json")
    densify_file(input_file, output_file)
    output = caplog.text
    expected_warning = r"WARNING .* unable to determine source CRS for file .*geometry-4326-no-crs.json, assumed CRS is OGC:CRS84\n"
    assert re.match(
        expected_warning, output
    ), f"stderr expected message is: {expected_warning}, actual message was: {output}"



import argparse
import logging
import os
import sys
from collections.abc import Callable
from enum import Enum
from functools import wraps
from typing import Any, Literal

from rich_argparse import RichHelpFormatter

from geodense import __version__, add_stderr_logger
from geodense.lib import (
    SUPPORTED_FILE_FORMATS,
    check_density_file,
    densify_file,
)
from geodense.models import DEFAULT_MAX_SEGMENT_LENGTH, GeodenseError

logger = logging.getLogger("geodense")


def cli_exception_handler(f: Callable) -> Callable:
    @wraps(f)
    def decorated(*args, **kwargs) -> Any:  # noqa: ANN002, ANN003, ANN401
        try:
            return f(*args, **kwargs)
        except GeodenseError as e:
            logger.error(e)
            sys.exit(1)
        except (
            Exception
        ) as e:  # unexpected exception, show stacktrace by calling logger.exception
            logger.exception(e)
            sys.exit(1)

    return decorated


@cli_exception_handler
def densify_cmd(  # noqa: PLR0913
    input_file: str,
    output_file: str,
    overwrite: bool = False,
    max_segment_length: float | None = None,
    in_projection: bool = False,
    src_crs: str | None = None,
) -> None:
    densify_file(
        input_file,
        output_file,
        overwrite,
        max_segment_length,
        in_projection,
        src_crs,
    )


@cli_exception_handler
def check_density_cmd(  # noqa: PLR0913
    input_file: str,
    max_segment_length: float,
    overwrite: bool = False,
    in_projection: bool = False,
    src_crs: str | None = None,
    density_check_report_path: str | None = None,
) -> None:
    print(overwrite)

    check_status, density_check_report_path, nr_line_segments = check_density_file(
        input_file,
        max_segment_length,
        density_check_report_path,
        src_crs,
        in_projection=in_projection,
        overwrite=overwrite,
    )

    status = "OK" if check_status else "FAILED"
    status_message = f"density-check {status} for file {input_file} with max-segment-length: {max_segment_length}"

    print(status_message)  # print status message for both OK and FAILED status

    if check_status:
        sys.exit(0)
    else:
        print(
            f"{nr_line_segments} linesegments in data exceeded max-segment-length {max_segment_length}, linesegment geometries written to GeoJSON FeatureCollection: {density_check_report_path}"
        )
        sys.exit(1)


def main() -> None:
    input_file_help = "any valid GeoJSON file, accepted GeoJSON objects: FeatureCollection, Feature, Geometry and GeometryCollection "
    source_crs_help = "override source CRS, if not specified then the CRS found in the GeoJSON input file will be used; format: $AUTH:$CODE; for example: EPSG:4326"
    verbose_help = "verbose output"
    max_segment_length_help = (
        f"max allowed segment length in meters; default: {DEFAULT_MAX_SEGMENT_LENGTH}"
    )

    parser = argparse.ArgumentParser(
        prog="geodense",
        description="Check density and densify geometries using the geodesic (ellipsoidal great-circle) calculation for accurate CRS transformations",
        epilog="Created by https://www.nsgi.nl/",
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-v", "--version", action="version", version=__version__)
    subparsers = parser.add_subparsers()

    densify_parser = subparsers.add_parser(
        "densify",
        formatter_class=parser.formatter_class,
        description="Densify (multi)polygon and (multi)linestring geometries along the geodesic (ellipsoidal great-circle), in base CRS (geographic) in case of projected source CRS.Supports GeoJSON as input file format. When supplying 3D coordinates, the height is linear interpolated for both geographic CRSs with ellipsoidal height and for compound CRSs with physical height.",
    )
    densify_parser.add_argument(
        "input_file",
        type=lambda x: is_json_file_arg(
            parser, x, "input_file", exist_required=FileRequired.exist
        ),
        help=input_file_help,
    )
    densify_parser.add_argument(
        "output_file",
        type=lambda x: is_json_file_arg(
            parser, x, "output_file", exist_required=FileRequired.either
        ),
        help="output file path",
    )

    densify_parser.add_argument(
        "--max-segment-length",
        "-m",
        type=float,
        default=DEFAULT_MAX_SEGMENT_LENGTH,
        help=max_segment_length_help,
    )

    densify_parser.add_argument(
        "--in-projection",
        "-p",
        action="store_true",
        default=False,
        help="densify using linear interpolation in source projection instead of the geodesic, not applicable when source CRS is geographic",
    )
    densify_parser.add_argument(
        "--overwrite",
        "-o",
        action="store_true",
        default=False,
        help="overwrite output file if exists",
    )

    densify_parser.add_argument(
        "-v", "--verbose", action="store_true", default=False, help=verbose_help
    )

    densify_parser.add_argument(
        "--src-crs",
        "-s",
        type=str,
        help=source_crs_help,
        default=None,
    )

    densify_parser.set_defaults(func=densify_cmd)

    check_density_parser = subparsers.add_parser(
        "check-density",
        formatter_class=parser.formatter_class,
        description="Check density of (multi)polygon and (multi)linestring geometries based on geodesic (ellipsoidal great-circle) distance, in base CRS (geographic) in case of projected source CRS. \
        When result of check is OK the program will return with exit code 0, when result \
        is FAILED the program will return with exit code 1. The density-check report is a GeoJSON FeatureCollection containing linesegments exceeding the max-segment-length treshold.",
    )
    check_density_parser.add_argument(
        "input_file",
        type=lambda x: is_json_file_arg(
            parser, x, "input_file", exist_required=FileRequired.exist
        ),
        help=input_file_help,
    )
    check_density_parser.add_argument(
        "--max-segment-length",
        "-m",
        type=float,
        default=DEFAULT_MAX_SEGMENT_LENGTH,
        help=max_segment_length_help,
    )
    check_density_parser.add_argument(
        "--src-crs",
        "-s",
        type=str,
        help=source_crs_help,
        default=None,
    )
    check_density_parser.add_argument(
        "--in-projection",
        "-p",
        action="store_true",
        default=False,
        help="check density using linear interpolation in source projection instead of the geodesic, not applicable when source CRS is geographic",
    )
    check_density_parser.add_argument(
        "--density-check-report-path",
        "-r",
        dest="density_check_report_path",
        required=False,
        help="density-check report path, when omitted a temp file will be used. Report is only generated when density-check fails.",
        metavar="FILE_PATH",
        type=lambda x: is_json_file_arg(
            parser, x, "density-check-report-path", FileRequired.either
        ),
    )
    check_density_parser.add_argument(
        "--overwrite",
        "-o",
        action="store_true",
        default=False,
        help="overwrite density-check report if exists",
    )
    check_density_parser.add_argument(
        "-v", "--verbose", action="store_true", default=False, help=verbose_help
    )
    check_density_parser.set_defaults(func=check_density_cmd)

    parser._positionals.title = "commands"
    args = parser.parse_args()

    try:
        add_stderr_logger(args.verbose)
        del args.verbose
        func = args.func
        del args.func
        func(**vars(args))
    except AttributeError as _:
        parser.print_help(file=sys.stderr)
        sys.exit(1)


class FileRequired(Enum):
    exist: Literal["exist"] = "exist"
    not_exist: Literal["not_exist"] = "not_exist"
    either: Literal["either"] = "either"


def is_json_file_arg(
    parser: argparse.ArgumentParser,
    arg: str,
    arg_name: str,
    exist_required: FileRequired,
) -> str:
    _, file_ext = os.path.splitext(arg)
    unsupported_file_extension_msg = "unsupported file extension of {input_file}, received: {ext}, expected one of: {supported_ext}"
    if arg != "-" and file_ext not in SUPPORTED_FILE_FORMATS["GeoJSON"]:
        parser.error(
            unsupported_file_extension_msg.format(
                input_file=arg_name,
                ext=file_ext,
                supported_ext=", ".join(SUPPORTED_FILE_FORMATS["GeoJSON"]),
            )
        )
    if (
        exist_required != FileRequired.either
        and arg != "-"
        and (
            not os.path.isfile(arg)
            if exist_required == FileRequired.exist
            else os.path.isfile(arg)
        )
    ):
        if exist_required:
            parser.error(f"{arg_name} {arg} does not exist")
        else:
            parser.error(f"{arg_name} {arg} exists")

    if (
        arg not in ["", "-"]
        and exist_required in [FileRequired.not_exist, FileRequired.either]
        and not os.path.exists(os.path.realpath(os.path.dirname(arg)))
    ):
        os.mkdir(os.path.realpath(os.path.dirname(arg)))
    return arg


if __name__ == "__main__":
    main()  # pragma: no cover



import re
from contextlib import nullcontext as does_not_raise
from typing import Any

import pytest
from _pytest.python_api import RaisesContext
from geodense.lib import _geom_type_check, _raise_e_if_point_geom
from geodense.models import GeodenseError
from geojson_pydantic import Feature


@pytest.mark.parametrize(
    ("geojson", "expectation"),
    [
        ("linestring_feature_gj", does_not_raise()),
        (
            "point_feature_gj",
            pytest.raises(
                GeodenseError,
                match=r"GeoJSON contains only \(Multi\)Point geometries",
            ),
        ),
        ("geometry_collection_gj", does_not_raise()),
    ],
)
def test_geom_type_check(
    geojson, expectation: Any | RaisesContext[GeodenseError], request
):
    with expectation:
        gj_obj = request.getfixturevalue(geojson)
        _geom_type_check(gj_obj)


def test_mixed_geom_outputs_warning(geometry_collection_feature_gj, caplog):
    geojson_obj = geometry_collection_feature_gj
    _geom_type_check(geojson_obj)
    my_regex = re.compile(r"WARNING .* GeoJSON contains \(Multi\)Point geometries\n")
    assert my_regex.match(caplog.text) is not None


@pytest.mark.parametrize(
    ("feature_fixture", "expected"),
    [
        (
            "point_feature_gj",
            pytest.raises(
                GeodenseError,
                match=r"received point geometry coordinates, instead of \(multi\)linestring",
            ),
        ),
        ("polygon_feature_with_holes_gj", does_not_raise()),
    ],
)
def test_raise_if_point(feature_fixture, expected, request):
    feature: Feature = request.getfixturevalue(feature_fixture)

    with expected:
        _raise_e_if_point_geom(feature.geometry.coordinates)



import os
import re
import tempfile
from contextlib import nullcontext as does_not_raise
from unittest import mock
from unittest.mock import MagicMock, patch

import geodense
import pytest
from cli_test_helpers import ArgvContext
from geodense.main import check_density_cmd, main
from geodense.models import DEFAULT_MAX_SEGMENT_LENGTH, GeodenseError


@patch("geodense.main.densify_cmd")
def test_cli_densify_cmd(mock_command, tmpdir, test_dir):
    in_file = "linestrings.json"
    in_filepath = f"{test_dir}/data/{in_file}"
    out_filepath = os.path.join(tmpdir, in_file)

    max_segment_length = "5000"
    with ArgvContext(
        "geodense",
        "densify",
        in_filepath,
        out_filepath,
        "--max-segment-length",
        max_segment_length,
        "--in-projection",
    ):
        main()

    assert mock_command.call_args.kwargs["input_file"] == in_filepath
    assert mock_command.call_args.kwargs["output_file"] == out_filepath
    assert mock_command.call_args.kwargs["max_segment_length"] == float(
        max_segment_length
    )  # note max_segment_length arg is parsed as int by argparse
    assert mock_command.call_args.kwargs["in_projection"] is True
    assert mock_command.called


@patch(
    "geodense.main.check_density_file",
    MagicMock(return_value=(True, "/tmp/bla/foobar.json", 0)),  # noqa: S108
)
def test_check_density_cmd_exit_0(test_dir):
    input_file = os.path.join(test_dir, "data", "polygons.json")
    with pytest.raises(SystemExit) as cm:
        check_density_cmd(input_file, 20000)
    assert cm.type == SystemExit
    expected_exit_code = 0
    assert (
        cm.value.code == expected_exit_code
    ), f"expected check_density_cmd call to exit with exit code {expected_exit_code} was {cm.value.code}"


@patch(
    "geodense.main.check_density_file", MagicMock(return_value=[([0, 1], 100.1239123)])
)
def test_check_density_cmd_exit_1_when_result_not_ok(test_dir):
    input_file = os.path.join(test_dir, "data", "polygons.json")
    with pytest.raises(SystemExit) as cm:
        check_density_cmd(input_file, 20000, "")
    assert cm.type == SystemExit
    expected_exit_code = 1
    assert (
        cm.value.code == expected_exit_code
    ), f"expected check_density_cmd call to exit with exit code {expected_exit_code} was {cm.value.code}"


@pytest.mark.parametrize(
    ("input_file", "output_file", "expectation"),
    [
        (
            "linestrings.foobar",
            "linestrings_out.foobar",
            (
                pytest.raises(SystemExit),
                2,
                r"(?sm).*geodense: error: unsupported file extension of input_file, received: \.foobar, expected one of: \.geojson, \.json.*",
            ),
        ),
        ("linestrings.json", "linestrings.geojson", (does_not_raise(), 0, None)),
    ],
)
def test_densify_file_unsupported_file_format(
    test_dir, input_file, output_file, expectation, capsys
):
    input_file = os.path.join(test_dir, "data", input_file)
    output_file = os.path.join(tempfile.mkdtemp(), output_file)

    expected_exit = expectation[0]
    with ArgvContext(
        "geodense", "densify", input_file, output_file
    ), expected_exit as cm:
        main()
    expected_exit_code = expectation[1]

    if expected_exit_code > 0:
        assert cm.type == SystemExit
        assert (
            cm.value.code == expected_exit_code
        ), f"expected check_density_cmd call to exit with exit code {expected_exit_code} was {cm.value.code}"

    expected_error_pattern = expectation[2]
    captured = capsys.readouterr()

    if expected_error_pattern is not None:
        assert re.match(
            expected_error_pattern,
            captured.err,
        )


@patch("geodense.main.check_density_cmd")
def test_cli_check_density_cmd(mock_command, test_dir):
    in_file = "linestrings.json"
    in_filepath = f"{test_dir}/data/{in_file}"

    max_segment_length = "5000"
    with ArgvContext(
        "geodense",
        "check-density",
        in_filepath,
        "--max-segment-length",
        max_segment_length,
    ):
        main()

    assert mock_command.call_args.kwargs["input_file"] == in_filepath
    assert mock_command.call_args.kwargs["max_segment_length"] == float(
        max_segment_length
    )  # note max_segment_length arg is parsed as int by argparse
    assert mock_command.called


def test_cli_densify_shows_outputs_error_returns_1(caplog, tmpdir, test_dir):
    with mock.patch.object(geodense.main, "densify_file") as get_mock:
        get_mock.side_effect = GeodenseError("FOOBAR")

        with pytest.raises(SystemExit) as cm, ArgvContext(
            "geodense",
            "densify",
            os.path.join(test_dir, "data", "linestrings.json"),
            os.path.join(tmpdir, "linestrings.json"),
        ):
            main()

        assert cm.type == SystemExit
        expected_exit_code = 1
        assert (
            cm.value.code == expected_exit_code
        ), f"expected densify_cmd call to exit with exit code {expected_exit_code} was {cm.value.code}"
        assert re.match(
            r"ERROR\s+geodense:main.py:.* FOOBAR\n",
            caplog.text,
        )


def test_cli_check_density_shows_outputs_error_returns_1(caplog, test_dir):
    with mock.patch.object(geodense.main, "check_density_file") as get_mock:
        get_mock.side_effect = GeodenseError("FOOBAR")

        with pytest.raises(SystemExit) as cm, ArgvContext(
            "geodense",
            "check-density",
            os.path.join(test_dir, "data", "linestrings.json"),
            "--max-segment-length",
            str(DEFAULT_MAX_SEGMENT_LENGTH),
        ):
            main()

        assert cm.type == SystemExit
        expected_exit_code = 1
        assert (
            cm.value.code == expected_exit_code
        ), f"expected check_density_cmd call to exit with exit code {expected_exit_code} was {cm.value.code}"
        assert re.match(
            r"ERROR\s+geodense:main.py:.* FOOBAR\n",
            caplog.text,
        )


USAGE_REGEX = r"^Usage: geodense (?:\[-[a-z]{1}\]\s)+\{.*?}"


def test_cli_shows_help_text_stderr_invoked_no_args(capsys):
    with pytest.raises(SystemExit), ArgvContext("geodense"):
        main()
    _, err = capsys.readouterr()
    assert re.match(USAGE_REGEX, err)


def test_cli_shows_help_text_invoked_help(capsys):
    with pytest.raises(SystemExit), ArgvContext("geodense", "--help"):
        main()
    out, _ = capsys.readouterr()
    assert re.match(USAGE_REGEX, out)
    assert "show this help message and exit" in out



import os
from unittest import mock

import pytest
from geodense.lib import (
    check_density_file,
    check_density_geometry_coordinates,
    flatten,
)
from geodense.models import DenseConfig, GeodenseError
from geodense.types import Nested, ReportLineString
from geojson_pydantic import Feature
from pyproj import CRS


def test_check_density(test_dir):
    result, _, nr_segments = check_density_file(
        os.path.join(test_dir, "data/linestrings.json"), 100
    )
    assert not result, "expected result is False"
    assert nr_segments > 0


def test_check_density_not_pass(linestring_feature_gj):
    feature: Feature = linestring_feature_gj
    result = []
    d_conf = DenseConfig(CRS.from_epsg(28992))
    result: Nested[ReportLineString] = check_density_geometry_coordinates(
        feature.geometry.coordinates, d_conf
    )
    flat_result: list[ReportLineString] = list(flatten(result))
    assert len(flat_result) > 0


def test_check_density_pass_linestring(linestring_feature_5000_gj):
    feature: Feature = linestring_feature_5000_gj
    result = []
    d_conf = DenseConfig(CRS.from_epsg(28992), 5000)
    check_density_geometry_coordinates(feature.geometry.coordinates, d_conf)
    assert len(result) == 0


def test_check_density_polygon_with_hole_not_pass(polygon_feature_with_holes_gj):
    feature: Feature = polygon_feature_with_holes_gj

    d_conf = DenseConfig(CRS.from_epsg(28992), 5000)

    result: Nested[ReportLineString] = check_density_geometry_coordinates(
        feature.geometry.coordinates, d_conf
    )
    flat_result: list[ReportLineString] = list(flatten(result))
    assert len(flat_result) > 0


def test_check_density_3d(linestring_3d_feature_gj):
    feature_t: Feature = linestring_3d_feature_gj
    d_conf = DenseConfig(CRS.from_epsg(7415), 500)
    result: Nested[ReportLineString] = check_density_geometry_coordinates(
        feature_t.geometry.coordinates, d_conf
    )
    flat_result: list[ReportLineString] = list(flatten(result))
    assert len(flat_result) > 0


@mock.patch("pyproj.Geod.inv", mock.MagicMock(return_value=(None, None, float("NaN"))))
def test_densify_file_exception(linestring_3d_feature_gj):
    feature_t: Feature = linestring_3d_feature_gj

    d_conf = DenseConfig(CRS.from_epsg(7415), 500)

    with pytest.raises(
        GeodenseError,
        match=r"unable to calculate geodesic distance, output calculation geodesic distance: nan, expected: floating-point number",
    ):
        _: Nested[ReportLineString] = check_density_geometry_coordinates(
            feature_t.geometry.coordinates, d_conf
        )



from collections.abc import Sequence
from typing import TypeAlias, TypeVar

from geojson_pydantic import (
    Feature,
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from geojson_pydantic.geometries import Geometry
from geojson_pydantic.types import (
    LineStringCoords,
    MultiLineStringCoords,
    MultiPointCoords,
    MultiPolygonCoords,
    Position,
)

from geodense.geojson import CrsFeatureCollection

T = TypeVar("T")

GeojsonGeomNoGeomCollection: TypeAlias = (
    Point | MultiPoint | LineString | MultiLineString | Polygon | MultiPolygon
)

GeojsonObject: TypeAlias = (
    Feature | CrsFeatureCollection | Geometry | GeometryCollection
)

GeojsonCoordinates: TypeAlias = (
    Position
    | MultiPointCoords
    | LineStringCoords
    | MultiLineStringCoords
    | MultiPolygonCoords
)


Nested: TypeAlias = Sequence[T | None | "Nested"]

ReportLineString = tuple[float, tuple[Position, Position]]
